# Copyright 2012 NetApp
# Copyright 2014 Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Unit tests for the Share driver module."""

import time

import ddt
import mock
from mock import PropertyMock

from manila import exception
from manila import network
from manila.share import configuration
from manila.share import driver
from manila import test
from manila.tests import utils as test_utils
from manila import utils


def fake_execute_with_raise(*cmd, **kwargs):
    raise exception.ProcessExecutionError


def fake_sleep(duration):
    pass


class ShareDriverWithExecuteMixin(driver.ShareDriver, driver.ExecuteMixin):
    pass


@ddt.ddt
class ShareDriverTestCase(test.TestCase):
    _SNAPSHOT_METHOD_NAMES = ["create_snapshot", "delete_snapshot"]

    def setUp(self):
        super(ShareDriverTestCase, self).setUp()
        self.utils = utils
        self.mock_object(self.utils, 'execute', fake_execute_with_raise)
        self.time = time
        self.mock_object(self.time, 'sleep', fake_sleep)
        driver.CONF.set_default('driver_handles_share_servers', True)

    def test__try_execute(self):
        execute_mixin = ShareDriverWithExecuteMixin(
            True, configuration=configuration.Configuration(None))
        self.assertRaises(exception.ProcessExecutionError,
                          execute_mixin._try_execute)

    def test_verify_share_driver_mode_option_type(self):
        data = {'DEFAULT': {'driver_handles_share_servers': 'True'}}
        with test_utils.create_temp_config_with_opts(data):
            share_driver = driver.ShareDriver([True, False])
            self.assertTrue(share_driver.driver_handles_share_servers)

    def _instantiate_share_driver(self, network_config_group,
                                  driver_handles_share_servers,
                                  admin_network_config_group=None):
        self.mock_object(network, 'API')
        config = mock.Mock()
        config.append_config_values = mock.Mock()
        config.config_group = 'fake_config_group'
        config.network_config_group = network_config_group
        if admin_network_config_group:
            config.admin_network_config_group = admin_network_config_group
        config.safe_get = mock.Mock(return_value=driver_handles_share_servers)

        share_driver = driver.ShareDriver([True, False], configuration=config)

        self.assertTrue(hasattr(share_driver, 'configuration'))
        config.append_config_values.assert_called_once_with(driver.share_opts)
        if driver_handles_share_servers:
            calls = []
            if network_config_group:
                calls.append(mock.call(
                    config_group_name=config.network_config_group))
            else:
                calls.append(mock.call(
                    config_group_name=config.config_group))
            if admin_network_config_group:
                calls.append(mock.call(
                    config_group_name=config.admin_network_config_group,
                    label='admin'))
            network.API.assert_has_calls(calls)
            self.assertTrue(hasattr(share_driver, 'network_api'))
            self.assertTrue(hasattr(share_driver, 'admin_network_api'))
            self.assertIsNotNone(share_driver.network_api)
            self.assertIsNotNone(share_driver.admin_network_api)
        else:
            self.assertFalse(hasattr(share_driver, 'network_api'))
            self.assertTrue(hasattr(share_driver, 'admin_network_api'))
            self.assertIsNone(share_driver.admin_network_api)
            self.assertFalse(network.API.called)
        return share_driver

    def test_instantiate_share_driver(self):
        self._instantiate_share_driver(None, True)

    def test_instantiate_share_driver_another_config_group(self):
        self._instantiate_share_driver("fake_network_config_group", True)

    def test_instantiate_share_driver_with_admin_network(self):
        self._instantiate_share_driver(
            "fake_network_config_group", True,
            "fake_admin_network_config_group")

    def test_instantiate_share_driver_no_configuration(self):
        self.mock_object(network, 'API')

        share_driver = driver.ShareDriver(True, configuration=None)

        self.assertIsNone(share_driver.configuration)
        network.API.assert_called_once_with(config_group_name=None)

    def test_get_share_stats_refresh_false(self):
        share_driver = driver.ShareDriver(True, configuration=None)
        share_driver._stats = {'fake_key': 'fake_value'}

        result = share_driver.get_share_stats(False)

        self.assertEqual(share_driver._stats, result)

    def test_get_share_stats_refresh_true(self):
        conf = configuration.Configuration(None)
        expected_keys = [
            'qos', 'driver_version', 'share_backend_name',
            'free_capacity_gb', 'total_capacity_gb',
            'driver_handles_share_servers',
            'reserved_percentage', 'vendor_name', 'storage_protocol',
            'snapshot_support', 'mount_snapshot_support',
        ]
        share_driver = driver.ShareDriver(True, configuration=conf)
        fake_stats = {'fake_key': 'fake_value'}
        share_driver._stats = fake_stats

        result = share_driver.get_share_stats(True)

        self.assertNotEqual(fake_stats, result)
        for key in expected_keys:
            self.assertIn(key, result)
        self.assertEqual('Open Source', result['vendor_name'])

    @ddt.data(
        {'opt': True, 'allowed': True},
        {'opt': True, 'allowed': (True, False)},
        {'opt': True, 'allowed': [True, False]},
        {'opt': True, 'allowed': set([True, False])},
        {'opt': False, 'allowed': False},
        {'opt': False, 'allowed': (True, False)},
        {'opt': False, 'allowed': [True, False]},
        {'opt': False, 'allowed': set([True, False])})
    @ddt.unpack
    def test__verify_share_server_handling_valid_cases(self, opt, allowed):
        conf = configuration.Configuration(None)
        self.mock_object(conf, 'safe_get', mock.Mock(return_value=opt))
        share_driver = driver.ShareDriver(allowed, configuration=conf)
        self.assertTrue(conf.safe_get.called)
        self.assertEqual(opt, share_driver.driver_handles_share_servers)

    @ddt.data(
        {'opt': False, 'allowed': True},
        {'opt': True, 'allowed': False},
        {'opt': None, 'allowed': True},
        {'opt': 'True', 'allowed': True},
        {'opt': 'False', 'allowed': False},
        {'opt': [], 'allowed': True},
        {'opt': True, 'allowed': []},
        {'opt': True, 'allowed': ['True']},
        {'opt': False, 'allowed': ['False']})
    @ddt.unpack
    def test__verify_share_server_handling_invalid_cases(self, opt, allowed):
        conf = configuration.Configuration(None)
        self.mock_object(conf, 'safe_get', mock.Mock(return_value=opt))
        self.assertRaises(
            exception.ManilaException,
            driver.ShareDriver, allowed, configuration=conf)
        self.assertTrue(conf.safe_get.called)

    def test_setup_server_handling_disabled(self):
        share_driver = self._instantiate_share_driver(None, False)
        # We expect successful execution, nothing to assert
        share_driver.setup_server('Nothing is expected to happen.')

    def test_setup_server_handling_enabled(self):
        share_driver = self._instantiate_share_driver(None, True)
        self.assertRaises(
            NotImplementedError,
            share_driver.setup_server,
            'fake_network_info')

    def test_teardown_server_handling_disabled(self):
        share_driver = self._instantiate_share_driver(None, False)
        # We expect successful execution, nothing to assert
        share_driver.teardown_server('Nothing is expected to happen.')

    def test_teardown_server_handling_enabled(self):
        share_driver = self._instantiate_share_driver(None, True)
        self.assertRaises(
            NotImplementedError,
            share_driver.teardown_server,
            'fake_share_server_details')

    def _assert_is_callable(self, obj, attr):
        self.assertTrue(callable(getattr(obj, attr)))

    @ddt.data('manage_existing',
              'unmanage')
    def test_drivers_methods_needed_by_manage_functionality(self, method):
        share_driver = self._instantiate_share_driver(None, False)

        self._assert_is_callable(share_driver, method)

    @ddt.data('manage_existing_snapshot',
              'unmanage_snapshot')
    def test_drivers_methods_needed_by_manage_snapshot_functionality(
            self, method):
        share_driver = self._instantiate_share_driver(None, False)

        self._assert_is_callable(share_driver, method)

    @ddt.data('revert_to_snapshot',
              'revert_to_replicated_snapshot')
    def test_drivers_methods_needed_by_share_revert_to_snapshot_functionality(
            self, method):
        share_driver = self._instantiate_share_driver(None, False)

        self._assert_is_callable(share_driver, method)

    @ddt.data(True, False)
    def test_get_share_server_pools(self, value):
        driver.CONF.set_default('driver_handles_share_servers', value)
        share_driver = driver.ShareDriver(value)
        self.assertEqual([],
                         share_driver.get_share_server_pools('fake_server'))

    @ddt.data(0.8, 1.0, 10.5, 20.0, None, '1', '1.1')
    def test_check_for_setup_error(self, value):
        driver.CONF.set_default('driver_handles_share_servers', False)
        share_driver = driver.ShareDriver(False)
        share_driver.configuration = configuration.Configuration(None)
        self.mock_object(share_driver.configuration, 'safe_get',
                         mock.Mock(return_value=value))
        if value and float(value) >= 1.0:
            share_driver.check_for_setup_error()
        else:
            self.assertRaises(exception.InvalidParameterValue,
                              share_driver.check_for_setup_error)

    def test_snapshot_support_exists(self):
        driver.CONF.set_default('driver_handles_share_servers', True)
        fake_method = lambda *args, **kwargs: None
        child_methods = {
            "create_snapshot": fake_method,
            "delete_snapshot": fake_method,
        }
        child_class_instance = type(
            "NotRedefined", (driver.ShareDriver, ), child_methods)(True)
        self.mock_object(child_class_instance, "configuration")

        child_class_instance._update_share_stats()

        self.assertTrue(child_class_instance._stats["snapshot_support"])
        self.assertTrue(child_class_instance.configuration.safe_get.called)

    @ddt.data(
        ([], [], False),
        (_SNAPSHOT_METHOD_NAMES, [], True),
        (_SNAPSHOT_METHOD_NAMES, _SNAPSHOT_METHOD_NAMES, True),
        (_SNAPSHOT_METHOD_NAMES[0:1], _SNAPSHOT_METHOD_NAMES[1:], True),
        ([], _SNAPSHOT_METHOD_NAMES, True),
    )
    @ddt.unpack
    def test_check_redefined_driver_methods(self, common_drv_meth_names,
                                            child_drv_meth_names,
                                            expected_result):
        # This test covers the case of drivers inheriting other drivers or
        # common classes.

        driver.CONF.set_default('driver_handles_share_servers', True)

        common_drv_methods, child_drv_methods = [
            {method_name: lambda *args, **kwargs: None
             for method_name in method_names}
            for method_names in (common_drv_meth_names,
                                 child_drv_meth_names)]

        common_drv = type(
            "NotRedefinedCommon", (driver.ShareDriver, ), common_drv_methods)
        child_drv_instance = type("NotRedefined", (common_drv, ),
                                  child_drv_methods)(True)

        has_redefined_methods = (
            child_drv_instance._has_redefined_driver_methods(
                self._SNAPSHOT_METHOD_NAMES))

        self.assertEqual(expected_result, has_redefined_methods)

    @ddt.data(
        (),
        ("create_snapshot"),
        ("delete_snapshot"),
        ("create_snapshot", "delete_snapshotFOO"),
    )
    def test_snapshot_support_absent(self, methods):
        driver.CONF.set_default('driver_handles_share_servers', True)
        fake_method = lambda *args, **kwargs: None
        child_methods = {}
        for method in methods:
            child_methods[method] = fake_method
        child_class_instance = type(
            "NotRedefined", (driver.ShareDriver, ), child_methods)(True)
        self.mock_object(child_class_instance, "configuration")

        child_class_instance._update_share_stats()

        self.assertFalse(child_class_instance._stats["snapshot_support"])
        self.assertTrue(child_class_instance.configuration.safe_get.called)

    @ddt.data(True, False)
    def test_snapshot_support_not_exists_and_set_explicitly(
            self, snapshots_are_supported):
        driver.CONF.set_default('driver_handles_share_servers', True)
        child_class_instance = type(
            "NotRedefined", (driver.ShareDriver, ), {})(True)
        self.mock_object(child_class_instance, "configuration")

        child_class_instance._update_share_stats(
            {"snapshot_support": snapshots_are_supported})

        self.assertEqual(
            snapshots_are_supported,
            child_class_instance._stats["snapshot_support"])
        self.assertTrue(child_class_instance.configuration.safe_get.called)

    @ddt.data(True, False)
    def test_snapshot_support_exists_and_set_explicitly(
            self, snapshots_are_supported):
        driver.CONF.set_default('driver_handles_share_servers', True)
        fake_method = lambda *args, **kwargs: None
        child_methods = {
            "create_snapshot": fake_method,
            "delete_snapshot": fake_method,
        }
        child_class_instance = type(
            "NotRedefined", (driver.ShareDriver, ), child_methods)(True)
        self.mock_object(child_class_instance, "configuration")

        child_class_instance._update_share_stats(
            {"snapshot_support": snapshots_are_supported})

        self.assertEqual(
            snapshots_are_supported,
            child_class_instance._stats["snapshot_support"])
        self.assertTrue(child_class_instance.configuration.safe_get.called)

    def test_create_share_from_snapshot_support_exists(self):
        driver.CONF.set_default('driver_handles_share_servers', True)
        fake_method = lambda *args, **kwargs: None
        child_methods = {
            "create_share_from_snapshot": fake_method,
            "create_snapshot": fake_method,
            "delete_snapshot": fake_method,
        }
        child_class_instance = type(
            "NotRedefined", (driver.ShareDriver, ), child_methods)(True)
        self.mock_object(child_class_instance, "configuration")

        child_class_instance._update_share_stats()

        self.assertTrue(
            child_class_instance._stats["create_share_from_snapshot_support"])
        self.assertTrue(child_class_instance.configuration.safe_get.called)

    @ddt.data(
        (),
        ("create_snapshot"),
        ("create_share_from_snapshotFOO"),
    )
    def test_create_share_from_snapshot_support_absent(self, methods):
        driver.CONF.set_default('driver_handles_share_servers', True)
        fake_method = lambda *args, **kwargs: None
        child_methods = {}
        for method in methods:
            child_methods[method] = fake_method
        child_class_instance = type(
            "NotRedefined", (driver.ShareDriver, ), child_methods)(True)
        self.mock_object(child_class_instance, "configuration")

        child_class_instance._update_share_stats()

        self.assertFalse(
            child_class_instance._stats["create_share_from_snapshot_support"])
        self.assertTrue(child_class_instance.configuration.safe_get.called)

    @ddt.data(True, False)
    def test_create_share_from_snapshot_not_exists_and_set_explicitly(
            self, creating_shares_from_snapshot_is_supported):
        driver.CONF.set_default('driver_handles_share_servers', True)
        child_class_instance = type(
            "NotRedefined", (driver.ShareDriver, ), {})(True)
        self.mock_object(child_class_instance, "configuration")

        child_class_instance._update_share_stats({
            "create_share_from_snapshot_support":
                creating_shares_from_snapshot_is_supported,
        })

        self.assertEqual(
            creating_shares_from_snapshot_is_supported,
            child_class_instance._stats["create_share_from_snapshot_support"])
        self.assertTrue(child_class_instance.configuration.safe_get.called)

    @ddt.data(True, False)
    def test_create_share_from_snapshot_exists_and_set_explicitly(
            self, create_share_from_snapshot_supported):
        driver.CONF.set_default('driver_handles_share_servers', True)
        fake_method = lambda *args, **kwargs: None
        child_methods = {"create_share_from_snapshot": fake_method}
        child_class_instance = type(
            "NotRedefined", (driver.ShareDriver, ), child_methods)(True)
        self.mock_object(child_class_instance, "configuration")

        child_class_instance._update_share_stats({
            "create_share_from_snapshot_support":
                create_share_from_snapshot_supported,
        })

        self.assertEqual(
            create_share_from_snapshot_supported,
            child_class_instance._stats["create_share_from_snapshot_support"])
        self.assertTrue(child_class_instance.configuration.safe_get.called)

    def test_get_periodic_hook_data(self):
        share_driver = self._instantiate_share_driver(None, False)
        share_instances = ["list", "of", "share", "instances"]

        result = share_driver.get_periodic_hook_data(
            "fake_context", share_instances)

        self.assertEqual(share_instances, result)

    def test_get_admin_network_allocations_number(self):
        share_driver = self._instantiate_share_driver(None, True)

        self.assertEqual(
            0, share_driver.get_admin_network_allocations_number())

    def test_allocate_admin_network_count_None(self):
        share_driver = self._instantiate_share_driver(None, True)
        ctxt = 'fake_context'
        share_server = 'fake_share_server'
        mock_get_admin_network_allocations_number = self.mock_object(
            share_driver,
            'get_admin_network_allocations_number',
            mock.Mock(return_value=0))
        self.mock_object(
            share_driver.admin_network_api,
            'allocate_network',
            mock.Mock(side_effect=Exception('ShouldNotBeRaised')))

        share_driver.allocate_admin_network(ctxt, share_server)

        mock_get_admin_network_allocations_number.assert_called_once_with()
        self.assertFalse(
            share_driver.admin_network_api.allocate_network.called)

    def test_allocate_admin_network_count_0(self):
        share_driver = self._instantiate_share_driver(None, True)
        ctxt = 'fake_context'
        share_server = 'fake_share_server'
        self.mock_object(
            share_driver,
            'get_admin_network_allocations_number',
            mock.Mock(return_value=0))
        self.mock_object(
            share_driver.admin_network_api,
            'allocate_network',
            mock.Mock(side_effect=Exception('ShouldNotBeRaised')))

        share_driver.allocate_admin_network(ctxt, share_server, count=0)

        self.assertFalse(
            share_driver.get_admin_network_allocations_number.called)
        self.assertFalse(
            share_driver.admin_network_api.allocate_network.called)

    def test_allocate_admin_network_count_1_api_initialized(self):
        share_driver = self._instantiate_share_driver(None, True)
        ctxt = 'fake_context'
        share_server = 'fake_share_server'
        mock_get_admin_network_allocations_number = self.mock_object(
            share_driver,
            'get_admin_network_allocations_number',
            mock.Mock(return_value=1))
        self.mock_object(
            share_driver.admin_network_api,
            'allocate_network',
            mock.Mock())

        share_driver.allocate_admin_network(ctxt, share_server)

        mock_get_admin_network_allocations_number.assert_called_once_with()
        (share_driver.admin_network_api.allocate_network.
            assert_called_once_with(ctxt, share_server, count=1))

    def test_allocate_admin_network_count_1_api_not_initialized(self):
        share_driver = self._instantiate_share_driver(None, True, None)
        ctxt = 'fake_context'
        share_server = 'fake_share_server'
        share_driver._admin_network_api = None
        mock_get_admin_network_allocations_number = self.mock_object(
            share_driver,
            'get_admin_network_allocations_number',
            mock.Mock(return_value=1))

        self.assertRaises(
            exception.NetworkBadConfigurationException,
            share_driver.allocate_admin_network,
            ctxt, share_server,
        )
        mock_get_admin_network_allocations_number.assert_called_once_with()

    def test_migration_start(self):

        driver.CONF.set_default('driver_handles_share_servers', False)
        share_driver = driver.ShareDriver(False)

        self.assertRaises(NotImplementedError, share_driver.migration_start,
                          None, None, None, None, None, None, None)

    def test_migration_continue(self):

        driver.CONF.set_default('driver_handles_share_servers', False)
        share_driver = driver.ShareDriver(False)

        self.assertRaises(NotImplementedError, share_driver.migration_continue,
                          None, None, None, None, None, None, None)

    def test_migration_complete(self):

        driver.CONF.set_default('driver_handles_share_servers', False)
        share_driver = driver.ShareDriver(False)

        self.assertRaises(NotImplementedError, share_driver.migration_complete,
                          None, None, None, None, None, None, None)

    def test_migration_cancel(self):

        driver.CONF.set_default('driver_handles_share_servers', False)
        share_driver = driver.ShareDriver(False)

        self.assertRaises(NotImplementedError, share_driver.migration_cancel,
                          None, None, None, None, None, None, None)

    def test_migration_get_progress(self):

        driver.CONF.set_default('driver_handles_share_servers', False)
        share_driver = driver.ShareDriver(False)

        self.assertRaises(NotImplementedError,
                          share_driver.migration_get_progress,
                          None, None, None, None, None, None, None)

    @ddt.data(True, False)
    def test_connection_get_info(self, admin):

        expected = {
            'mount': 'mount -vt nfs %(options)s /fake/fake_id %(path)s',
            'unmount': 'umount -v %(path)s',
            'access_mapping': {
                'ip': ['nfs']
            }
        }

        fake_share = {
            'id': 'fake_id',
            'share_proto': 'nfs',
            'export_locations': [{
                'path': '/fake/fake_id',
                'is_admin_only': admin
            }]
        }

        driver.CONF.set_default('driver_handles_share_servers', False)
        share_driver = driver.ShareDriver(False)
        share_driver.configuration = configuration.Configuration(None)

        connection_info = share_driver.connection_get_info(
            None, fake_share, "fake_server")

        self.assertEqual(expected, connection_info)

    def test_migration_check_compatibility(self):

        driver.CONF.set_default('driver_handles_share_servers', False)
        share_driver = driver.ShareDriver(False)
        share_driver.configuration = configuration.Configuration(None)
        expected = {
            'compatible': False,
            'writable': False,
            'preserve_metadata': False,
            'nondisruptive': False,
            'preserve_snapshots': False,
        }

        result = share_driver.migration_check_compatibility(
            None, None, None, None, None)

        self.assertEqual(expected, result)

    def test_update_access(self):
        share_driver = driver.ShareDriver(True, configuration=None)
        self.assertRaises(
            NotImplementedError,
            share_driver.update_access,
            'ctx',
            'fake_share',
            'fake_access_rules',
            'fake_add_rules',
            'fake_delete_rules'
        )

    def test_create_replica(self):
        share_driver = self._instantiate_share_driver(None, True)
        self.assertRaises(NotImplementedError,
                          share_driver.create_replica,
                          'fake_context', ['r1', 'r2'],
                          'fake_new_replica', [], [])

    def test_delete_replica(self):
        share_driver = self._instantiate_share_driver(None, True)
        self.assertRaises(NotImplementedError,
                          share_driver.delete_replica,
                          'fake_context', ['r1', 'r2'],
                          'fake_replica', [])

    def test_promote_replica(self):
        share_driver = self._instantiate_share_driver(None, True)
        self.assertRaises(NotImplementedError,
                          share_driver.promote_replica,
                          'fake_context', [], 'fake_replica', [])

    def test_update_replica_state(self):
        share_driver = self._instantiate_share_driver(None, True)
        self.assertRaises(NotImplementedError,
                          share_driver.update_replica_state,
                          'fake_context', ['r1', 'r2'], 'fake_replica', [], [])

    def test_create_replicated_snapshot(self):
        share_driver = self._instantiate_share_driver(None, False)
        self.assertRaises(NotImplementedError,
                          share_driver.create_replicated_snapshot,
                          'fake_context', ['r1', 'r2'], ['s1', 's2'])

    def test_delete_replicated_snapshot(self):
        share_driver = self._instantiate_share_driver(None, False)
        self.assertRaises(NotImplementedError,
                          share_driver.delete_replicated_snapshot,
                          'fake_context', ['r1', 'r2'], ['s1', 's2'])

    def test_update_replicated_snapshot(self):
        share_driver = self._instantiate_share_driver(None, False)
        self.assertRaises(NotImplementedError,
                          share_driver.update_replicated_snapshot,
                          'fake_context', ['r1', 'r2'], 'r1',
                          ['s1', 's2'], 's1')

    @ddt.data(True, False)
    def test_share_group_snapshot_support_exists_and_equals_snapshot_support(
            self, snapshots_are_supported):
        driver.CONF.set_default('driver_handles_share_servers', True)
        child_class_instance = driver.ShareDriver(True)
        child_class_instance._snapshots_are_supported = snapshots_are_supported
        self.mock_object(child_class_instance, "configuration")

        child_class_instance._update_share_stats()

        self.assertEqual(
            snapshots_are_supported,
            child_class_instance._stats["snapshot_support"])
        self.assertTrue(child_class_instance.configuration.safe_get.called)

    def test_create_share_group_from_share_group_snapshot(self):
        share_driver = self._instantiate_share_driver(None, False)
        fake_shares = [
            {'id': 'fake_share_%d' % i,
             'source_share_group_snapshot_member_id': 'fake_member_%d' % i}
            for i in (1, 2)]
        fake_share_group_dict = {
            'source_share_group_snapshot_id': 'some_fake_uuid_abc',
            'shares': fake_shares,
            'id': 'some_fake_uuid_def',
        }
        fake_share_group_snapshot_dict = {
            'share_group_snapshot_members': [
                {'id': 'fake_member_1'}, {'id': 'fake_member_2'}],
            'id': 'fake_share_group_snapshot_id',
        }
        mock_create = self.mock_object(
            share_driver, 'create_share_from_snapshot',
            mock.Mock(side_effect=['fake_export1', 'fake_export2']))
        expected_share_updates = [
            {
                'id': 'fake_share_1',
                'export_locations': 'fake_export1',
            },
            {
                'id': 'fake_share_2',
                'export_locations': 'fake_export2',
            },
        ]

        share_group_update, share_update = (
            share_driver.create_share_group_from_share_group_snapshot(
                'fake_context', fake_share_group_dict,
                fake_share_group_snapshot_dict))

        mock_create.assert_has_calls([
            mock.call(
                'fake_context',
                {'id': 'fake_share_1',
                 'source_share_group_snapshot_member_id': 'fake_member_1'},
                {'id': 'fake_member_1'}),
            mock.call(
                'fake_context',
                {'id': 'fake_share_2',
                 'source_share_group_snapshot_member_id': 'fake_member_2'},
                {'id': 'fake_member_2'})
        ])
        self.assertIsNone(share_group_update)
        self.assertEqual(expected_share_updates, share_update)

    def test_create_share_group_from_share_group_snapshot_dhss(self):
        share_driver = self._instantiate_share_driver(None, True)
        mock_share_server = mock.Mock()
        fake_shares = [
            {'id': 'fake_share_1',
             'source_share_group_snapshot_member_id': 'foo_member_1'},
            {'id': 'fake_share_2',
             'source_share_group_snapshot_member_id': 'foo_member_2'}]
        fake_share_group_dict = {
            'source_share_group_snapshot_id': 'some_fake_uuid',
            'shares': fake_shares,
            'id': 'eda52174-0442-476d-9694-a58327466c14',
        }
        fake_share_group_snapshot_dict = {
            'share_group_snapshot_members': [
                {'id': 'foo_member_1'}, {'id': 'foo_member_2'}],
            'id': 'fake_share_group_snapshot_id'
        }
        mock_create = self.mock_object(
            share_driver, 'create_share_from_snapshot',
            mock.Mock(side_effect=['fake_export1', 'fake_export2']))
        expected_share_updates = [
            {'id': 'fake_share_1', 'export_locations': 'fake_export1'},
            {'id': 'fake_share_2', 'export_locations': 'fake_export2'},
        ]

        share_group_update, share_update = (
            share_driver.create_share_group_from_share_group_snapshot(
                'fake_context',
                fake_share_group_dict,
                fake_share_group_snapshot_dict, share_server=mock_share_server,
            )
        )

        mock_create.assert_has_calls([
            mock.call(
                'fake_context',
                {'id': 'fake_share_%d' % i,
                 'source_share_group_snapshot_member_id': 'foo_member_%d' % i},
                {'id': 'foo_member_%d' % i},
                share_server=mock_share_server)
            for i in (1, 2)
        ])
        self.assertIsNone(share_group_update)
        self.assertEqual(expected_share_updates, share_update)

    def test_create_share_group_from_sg_snapshot_with_no_members(self):
        share_driver = self._instantiate_share_driver(None, False)
        fake_share_group_dict = {}
        fake_share_group_snapshot_dict = {'share_group_snapshot_members': []}

        share_group_update, share_update = (
            share_driver.create_share_group_from_share_group_snapshot(
                'fake_context', fake_share_group_dict,
                fake_share_group_snapshot_dict))

        self.assertIsNone(share_group_update)
        self.assertIsNone(share_update)

    def test_create_share_group_snapshot(self):
        fake_snap_member_1 = {
            'id': '6813e06b-a8f5-4784-b17d-f3e91afa370e',
            'share_id': 'a3ebdba5-b4e1-46c8-a0ea-a9ac8daf5296',
            'share_group_snapshot_id': 'fake_share_group_snapshot_id',
            'share_instance_id': 'fake_share_instance_id_1',
            'provider_location': 'should_not_be_used_1',
            'share': {
                'id': '420f978b-dbf6-4b3c-92fe-f5b17a0bb5e2',
                'size': 3,
                'share_proto': 'fake_share_proto',
            },
        }
        fake_snap_member_2 = {
            'id': '1e010dfe-545b-432d-ab95-4ef03cd82f89',
            'share_id': 'a3ebdba5-b4e1-46c8-a0ea-a9ac8daf5296',
            'share_group_snapshot_id': 'fake_share_group_snapshot_id',
            'share_instance_id': 'fake_share_instance_id_2',
            'provider_location': 'should_not_be_used_2',
            'share': {
                'id': '420f978b-dbf6-4b3c-92fe-f5b17a0bb5e2',
                'size': '2',
                'share_proto': 'fake_share_proto',
            },
        }
        fake_snap_dict = {
            'status': 'available',
            'project_id': '13c0be6290934bd98596cfa004650049',
            'user_id': 'a0314a441ca842019b0952224aa39192',
            'description': None,
            'deleted': '0',
            'share_group_id': '4b04fdc3-00b9-4909-ba1a-06e9b3f88b67',
            'share_group_snapshot_members': [
                fake_snap_member_1, fake_snap_member_2],
            'deleted_at': None,
            'id': 'f6aa3b59-57eb-421e-965c-4e182538e36a',
            'name': None
        }
        share_driver = self._instantiate_share_driver(None, False)
        share_driver._stats['snapshot_support'] = True
        mock_create_snap = self.mock_object(
            share_driver, 'create_snapshot',
            mock.Mock(side_effect=lambda *args, **kwargs: {
                'foo_k': 'foo_v', 'bar_k': 'bar_v_%s' % args[1]['id']}))

        share_group_snapshot_update, member_update_list = (
            share_driver.create_share_group_snapshot(
                'fake_context', fake_snap_dict))

        mock_create_snap.assert_has_calls([
            mock.call(
                'fake_context',
                {'snapshot_id': member['share_group_snapshot_id'],
                 'share_id': member['share_id'],
                 'share_instance_id': member['share']['id'],
                 'id': member['id'],
                 'share': member['share'],
                 'size': member['share']['size'],
                 'share_size': member['share']['size'],
                 'share_proto': member['share']['share_proto'],
                 'provider_location': None},
                share_server=None)
            for member in (fake_snap_member_1, fake_snap_member_2)
        ])
        self.assertIsNone(share_group_snapshot_update)
        self.assertEqual(
            [{'id': member['id'], 'foo_k': 'foo_v',
              'bar_k': 'bar_v_%s' % member['id']}
             for member in (fake_snap_member_1, fake_snap_member_2)],
            member_update_list,
        )

    def test_create_share_group_snapshot_failed_snapshot(self):
        fake_snap_member_1 = {
            'id': '6813e06b-a8f5-4784-b17d-f3e91afa370e',
            'share_id': 'a3ebdba5-b4e1-46c8-a0ea-a9ac8daf5296',
            'share_group_snapshot_id': 'fake_share_group_snapshot_id',
            'share_instance_id': 'fake_share_instance_id_1',
            'provider_location': 'should_not_be_used_1',
            'share': {
                'id': '420f978b-dbf6-4b3c-92fe-f5b17a0bb5e2',
                'size': 3,
                'share_proto': 'fake_share_proto',
            },
        }
        fake_snap_member_2 = {
            'id': '1e010dfe-545b-432d-ab95-4ef03cd82f89',
            'share_id': 'a3ebdba5-b4e1-46c8-a0ea-a9ac8daf5296',
            'share_group_snapshot_id': 'fake_share_group_snapshot_id',
            'share_instance_id': 'fake_share_instance_id_2',
            'provider_location': 'should_not_be_used_2',
            'share': {
                'id': '420f978b-dbf6-4b3c-92fe-f5b17a0bb5e2',
                'size': '2',
                'share_proto': 'fake_share_proto',
            },
        }
        fake_snap_dict = {
            'status': 'available',
            'project_id': '13c0be6290934bd98596cfa004650049',
            'user_id': 'a0314a441ca842019b0952224aa39192',
            'description': None,
            'deleted': '0',
            'share_group_id': '4b04fdc3-00b9-4909-ba1a-06e9b3f88b67',
            'share_group_snapshot_members': [
                fake_snap_member_1, fake_snap_member_2],
            'deleted_at': None,
            'id': 'f6aa3b59-57eb-421e-965c-4e182538e36a',
            'name': None
        }
        expected_exception = exception.ManilaException

        share_driver = self._instantiate_share_driver(None, False)
        share_driver._stats['snapshot_support'] = True
        mock_create_snap = self.mock_object(
            share_driver, 'create_snapshot',
            mock.Mock(side_effect=[None, expected_exception]))
        mock_delete_snap = self.mock_object(share_driver, 'delete_snapshot')

        self.assertRaises(
            expected_exception,
            share_driver.create_share_group_snapshot,
            'fake_context', fake_snap_dict)

        fake_snap_member_1_expected = {
            'snapshot_id': fake_snap_member_1['share_group_snapshot_id'],
            'share_id': fake_snap_member_1['share_id'],
            'share_instance_id': fake_snap_member_1['share']['id'],
            'id': fake_snap_member_1['id'],
            'share': fake_snap_member_1['share'],
            'size': fake_snap_member_1['share']['size'],
            'share_size': fake_snap_member_1['share']['size'],
            'share_proto': fake_snap_member_1['share']['share_proto'],
            'provider_location': None,
        }
        mock_create_snap.assert_has_calls([
            mock.call(
                'fake_context',
                {'snapshot_id': member['share_group_snapshot_id'],
                 'share_id': member['share_id'],
                 'share_instance_id': member['share']['id'],
                 'id': member['id'],
                 'share': member['share'],
                 'size': member['share']['size'],
                 'share_size': member['share']['size'],
                 'share_proto': member['share']['share_proto'],
                 'provider_location': None},
                share_server=None)
            for member in (fake_snap_member_1, fake_snap_member_2)
        ])
        mock_delete_snap.assert_called_with(
            'fake_context', fake_snap_member_1_expected, share_server=None)

    def test_create_share_group_snapshot_no_support(self):
        fake_snap_dict = {
            'status': 'available',
            'project_id': '13c0be6290934bd98596cfa004650049',
            'user_id': 'a0314a441ca842019b0952224aa39192',
            'description': None,
            'deleted': '0',
            'share_group_id': '4b04fdc3-00b9-4909-ba1a-06e9b3f88b67',
            'share_group_snapshot_members': [
                {
                    'status': 'available',
                    'share_type_id': '1a9ed31e-ee70-483d-93ba-89690e028d7f',
                    'user_id': 'a0314a441ca842019b0952224aa39192',
                    'deleted': 'False',
                    'share_proto': 'NFS',
                    'project_id': '13c0be6290934bd98596cfa004650049',
                    'share_group_snapshot_id':
                        'f6aa3b59-57eb-421e-965c-4e182538e36a',
                    'deleted_at': None,
                    'id': '6813e06b-a8f5-4784-b17d-f3e91afa370e',
                    'size': 1
                },
            ],
            'deleted_at': None,
            'id': 'f6aa3b59-57eb-421e-965c-4e182538e36a',
            'name': None
        }
        share_driver = self._instantiate_share_driver(None, False)
        share_driver._stats['snapshot_support'] = False

        self.assertRaises(
            exception.ShareGroupSnapshotNotSupported,
            share_driver.create_share_group_snapshot,
            'fake_context', fake_snap_dict)

    def test_create_share_group_snapshot_no_members(self):
        fake_snap_dict = {
            'status': 'available',
            'project_id': '13c0be6290934bd98596cfa004650049',
            'user_id': 'a0314a441ca842019b0952224aa39192',
            'description': None,
            'deleted': '0',
            'share_group_id': '4b04fdc3-00b9-4909-ba1a-06e9b3f88b67',
            'share_group_snapshot_members': [],
            'deleted_at': None,
            'id': 'f6aa3b59-57eb-421e-965c-4e182538e36a',
            'name': None
        }
        share_driver = self._instantiate_share_driver(None, False)
        share_driver._stats['snapshot_support'] = True

        share_group_snapshot_update, member_update_list = (
            share_driver.create_share_group_snapshot(
                'fake_context', fake_snap_dict))

        self.assertIsNone(share_group_snapshot_update)
        self.assertIsNone(member_update_list)

    def test_delete_share_group_snapshot(self):
        fake_snap_member_1 = {
            'id': '6813e06b-a8f5-4784-b17d-f3e91afa370e',
            'share_id': 'a3ebdba5-b4e1-46c8-a0ea-a9ac8daf5296',
            'share_group_snapshot_id': 'fake_share_group_snapshot_id',
            'share_instance_id': 'fake_share_instance_id_1',
            'provider_location': 'fake_provider_location_2',
            'share': {
                'id': '420f978b-dbf6-4b3c-92fe-f5b17a0bb5e2',
                'size': 3,
                'share_proto': 'fake_share_proto',
            },
        }
        fake_snap_member_2 = {
            'id': '1e010dfe-545b-432d-ab95-4ef03cd82f89',
            'share_id': 'a3ebdba5-b4e1-46c8-a0ea-a9ac8daf5296',
            'share_group_snapshot_id': 'fake_share_group_snapshot_id',
            'share_instance_id': 'fake_share_instance_id_2',
            'provider_location': 'fake_provider_location_2',
            'share': {
                'id': '420f978b-dbf6-4b3c-92fe-f5b17a0bb5e2',
                'size': '2',
                'share_proto': 'fake_share_proto',
            },
        }
        fake_snap_dict = {
            'status': 'available',
            'project_id': '13c0be6290934bd98596cfa004650049',
            'user_id': 'a0314a441ca842019b0952224aa39192',
            'description': None,
            'deleted': '0',
            'share_group_id': '4b04fdc3-00b9-4909-ba1a-06e9b3f88b67',
            'share_group_snapshot_members': [
                fake_snap_member_1, fake_snap_member_2],
            'deleted_at': None,
            'id': 'f6aa3b59-57eb-421e-965c-4e182538e36a',
            'name': None
        }

        share_driver = self._instantiate_share_driver(None, False)
        share_driver._stats['share_group_snapshot_support'] = True
        mock_delete_snap = self.mock_object(share_driver, 'delete_snapshot')

        share_group_snapshot_update, member_update_list = (
            share_driver.delete_share_group_snapshot(
                'fake_context', fake_snap_dict))

        mock_delete_snap.assert_has_calls([
            mock.call(
                'fake_context',
                {'snapshot_id': member['share_group_snapshot_id'],
                 'share_id': member['share_id'],
                 'share_instance_id': member['share']['id'],
                 'id': member['id'],
                 'share': member['share'],
                 'size': member['share']['size'],
                 'share_size': member['share']['size'],
                 'share_proto': member['share']['share_proto'],
                 'provider_location': member['provider_location']},
                share_server=None)
            for member in (fake_snap_member_1, fake_snap_member_2)
        ])
        self.assertIsNone(share_group_snapshot_update)
        self.assertIsNone(member_update_list)

    def test_snapshot_update_access(self):
        share_driver = self._instantiate_share_driver(None, False)
        self.assertRaises(NotImplementedError,
                          share_driver.snapshot_update_access,
                          'fake_context', 'fake_snapshot', ['r1', 'r2'],
                          [], [])

    @ddt.data({'user_networks': set([4]), 'conf': [4],
               'expected': {'ipv4': True, 'ipv6': False}},
              {'user_networks': set([6]), 'conf': [4],
               'expected': {'ipv4': False, 'ipv6': False}},
              {'user_networks': set([4, 6]), 'conf': [4],
               'expected': {'ipv4': True, 'ipv6': False}},
              {'user_networks': set([4]), 'conf': [6],
               'expected': {'ipv4': False, 'ipv6': False}},
              {'user_networks': set([6]), 'conf': [6],
               'expected': {'ipv4': False, 'ipv6': True}},
              {'user_networks': set([4, 6]), 'conf': [6],
               'expected': {'ipv4': False, 'ipv6': True}},
              {'user_networks': set([4]), 'conf': [4, 6],
               'expected': {'ipv4': True, 'ipv6': False}},
              {'user_networks': set([6]), 'conf': [4, 6],
               'expected': {'ipv4': False, 'ipv6': True}},
              {'user_networks': set([4, 6]), 'conf': [4, 6],
               'expected': {'ipv4': True, 'ipv6': True}},
              )
    @ddt.unpack
    def test_add_ip_version_capability_if_dhss_true(self,
                                                    user_networks,
                                                    conf,
                                                    expected):
        share_driver = self._instantiate_share_driver(None, True)
        self.mock_object(share_driver, 'get_configured_ip_versions',
                         mock.Mock(return_value=conf))
        versions = PropertyMock(return_value=user_networks)
        type(share_driver.network_api).enabled_ip_versions = versions
        data = {'share_backend_name': 'fake_backend'}

        result = share_driver.add_ip_version_capability(data)

        self.assertIsNotNone(result['ipv4_support'])
        self.assertEqual(expected['ipv4'], result['ipv4_support'])
        self.assertIsNotNone(result['ipv6_support'])
        self.assertEqual(expected['ipv6'], result['ipv6_support'])

    @ddt.data({'conf': [4],
               'expected': {'ipv4': True, 'ipv6': False}},
              {'conf': [6],
               'expected': {'ipv4': False, 'ipv6': True}},
              {'conf': [4, 6],
               'expected': {'ipv4': True, 'ipv6': True}},
              )
    @ddt.unpack
    def test_add_ip_version_capability_if_dhss_false(self, conf, expected):
        share_driver = self._instantiate_share_driver(None, False)
        self.mock_object(share_driver, 'get_configured_ip_versions',
                         mock.Mock(return_value=conf))
        data = {'share_backend_name': 'fake_backend'}
        result = share_driver.add_ip_version_capability(data)

        self.assertIsNotNone(result['ipv4_support'])
        self.assertEqual(expected['ipv4'], result['ipv4_support'])
        self.assertIsNotNone(result['ipv6_support'])
        self.assertEqual(expected['ipv6'], result['ipv6_support'])



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_policy import policy

from manila.policies import base


BASE_POLICY_NAME = 'share:%s'

# These deprecated rules can be removed in the 'Train' release.
deprecated_create_public_share_rule = policy.DeprecatedRule(
    name=BASE_POLICY_NAME % 'create_public_share',
    check_str=base.RULE_DEFAULT,
)

deprecated_set_public_share_rule = policy.DeprecatedRule(
    name=BASE_POLICY_NAME % 'set_public_share',
    check_str=base.RULE_DEFAULT,
)


shares_policies = [
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'create',
        check_str="",
        description="Create share.",
        operations=[
            {
                'method': 'POST',
                'path': '/shares',
            }
        ]),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'create_public_share',
        check_str=base.RULE_ADMIN_API,
        description="Create shares visible across all projects in the cloud. "
                    "This option will default to rule:admin_api in the "
                    "9.0.0 (Train) release of the OpenStack Shared File "
                    "Systems (manila) service.",
        deprecated_rule=deprecated_create_public_share_rule,
        deprecated_reason="Public shares must be accessible across the "
                          "cloud, irrespective of project namespaces. To "
                          "avoid unintended consequences, rule:admin_api "
                          "serves as a better default for this policy.",
        deprecated_since='S',
        operations=[
            {
                'method': 'POST',
                'path': '/shares',
            }
        ]),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'get',
        check_str=base.RULE_DEFAULT,
        description="Get share.",
        operations=[
            {
                'method': 'GET',
                'path': '/shares/{share_id}',
            }
        ]),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'get_all',
        check_str=base.RULE_DEFAULT,
        description="List shares.",
        operations=[
            {
                'method': 'GET',
                'path': '/shares',
            },
            {
                'method': 'GET',
                'path': '/shares/detail',
            }
        ]),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'update',
        check_str=base.RULE_DEFAULT,
        description="Update share.",
        operations=[
            {
                'method': 'PUT',
                'path': '/shares',
            }
        ]),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'set_public_share',
        check_str=base.RULE_ADMIN_API,
        description="Update shares to be visible across all projects in the "
                    "cloud. This option will default to rule:admin_api in the "
                    "9.0.0 (Train) release of the OpenStack Shared File "
                    "Systems (manila) service.",
        deprecated_rule=deprecated_set_public_share_rule,
        deprecated_reason="Public shares must be accessible across the "
                          "cloud, irrespective of project namespaces. To "
                          "avoid unintended consequences, rule:admin_api "
                          "serves as a better default for this policy.",
        deprecated_since='S',
        operations=[
            {
                'method': 'PUT',
                'path': '/shares',
            }
        ]),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'delete',
        check_str=base.RULE_DEFAULT,
        description="Delete share.",
        operations=[
            {
                'method': 'DELETE',
                'path': '/shares/{share_id}',
            }
        ]),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'force_delete',
        check_str=base.RULE_ADMIN_API,
        description="Force Delete a share.",
        operations=[
            {
                'method': 'DELETE',
                'path': '/shares/{share_id}',
            }
        ]),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'manage',
        check_str=base.RULE_ADMIN_API,
        description="Manage share.",
        operations=[
            {
                'method': 'POST',
                'path': '/shares/manage',
            }
        ]),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'unmanage',
        check_str=base.RULE_ADMIN_API,
        description="Unmanage share.",
        operations=[
            {
                'method': 'POST',
                'path': '/shares/unmanage',
            }
        ]),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'create_snapshot',
        check_str=base.RULE_DEFAULT,
        description="Create share snapshot.",
        operations=[
            {
                'method': 'POST',
                'path': '/snapshots',
            }
        ]),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'list_by_host',
        check_str=base.RULE_ADMIN_API,
        description="List share by host.",
        operations=[
            {
                'method': 'GET',
                'path': '/shares',
            },
            {
                'method': 'GET',
                'path': '/shares/detail',
            }
        ]),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'list_by_share_server_id',
        check_str=base.RULE_ADMIN_API,
        description="List share by server id.",
        operations=[
            {
                'method': 'GET',
                'path': '/shares'
            },
            {
                'method': 'GET',
                'path': '/shares/detail',
            }
        ]),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'access_get',
        check_str=base.RULE_DEFAULT,
        description="Get share access rule, it under deny access operation.",
        operations=[
            {
                'method': 'POST',
                'path': '/shares/{share_id}/action',
            }
        ]),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'access_get_all',
        check_str=base.RULE_DEFAULT,
        description="List share access rules.",
        operations=[
            {
                'method': 'GET',
                'path': '/shares/{share_id}/action',
            }
        ]),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'extend',
        check_str=base.RULE_DEFAULT,
        description="Extend share.",
        operations=[
            {
                'method': 'POST',
                'path': '/shares/{share_id}/action',
            }
        ]),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'shrink',
        check_str=base.RULE_DEFAULT,
        description="Shrink share.",
        operations=[
            {
                'method': 'POST',
                'path': '/shares/{share_id}/action',
            }
        ]),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'migration_start',
        check_str=base.RULE_ADMIN_API,
        description="Migrate a share to the specified host.",
        operations=[
            {
                'method': 'POST',
                'path': '/shares/{share_id}/action',
            }
        ]),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'migration_complete',
        check_str=base.RULE_ADMIN_API,
        description="Invokes 2nd phase of share migration.",
        operations=[
            {
                'method': 'POST',
                'path': '/shares/{share_id}/action',
            }
        ]),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'migration_cancel',
        check_str=base.RULE_ADMIN_API,
        description="Attempts to cancel share migration.",
        operations=[
            {
                'method': 'POST',
                'path': '/shares/{share_id}/action',
            }
        ]),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'migration_get_progress',
        check_str=base.RULE_ADMIN_API,
        description=("Retrieve share migration progress for a given "
                     "share."),
        operations=[
            {
                'method': 'POST',
                'path': '/shares/{share_id}/action',
            }
        ]),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'reset_task_state',
        check_str=base.RULE_ADMIN_API,
        description=("Reset task state."),
        operations=[
            {
                'method': 'POST',
                'path': '/shares/{share_id}/action',
            }
        ]),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'reset_status',
        check_str=base.RULE_ADMIN_API,
        description=("Reset status."),
        operations=[
            {
                'method': 'POST',
                'path': '/shares/{share_id}/action',
            }
        ]),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'revert_to_snapshot',
        check_str=base.RULE_DEFAULT,
        description=("Revert a share to a snapshot."),
        operations=[
            {
                'method': 'POST',
                'path': '/shares/{share_id}/action',
            }
        ]),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'allow_access',
        check_str=base.RULE_DEFAULT,
        description=("Add share access rule."),
        operations=[
            {
                'method': 'POST',
                'path': '/shares/{share_id}/action',
            }
        ]),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'deny_access',
        check_str=base.RULE_DEFAULT,
        description=("Remove share access rule."),
        operations=[
            {
                'method': 'POST',
                'path': '/shares/{share_id}/action',
            }
        ]),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'delete_snapshot',
        check_str=base.RULE_DEFAULT,
        description=("Delete share snapshot."),
        operations=[
            {
                'method': 'DELETE',
                'path': '/snapshots/{snapshot_id}',
            }
        ]),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'snapshot_update',
        check_str=base.RULE_DEFAULT,
        description=("Update share snapshot."),
        operations=[
            {
                'method': 'PUT',
                'path': '/snapshots/{snapshot_id}/action',
            }
        ]),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'update_share_metadata',
        check_str=base.RULE_DEFAULT,
        description=("Update share metadata."),
        operations=[
            {
                'method': 'PUT',
                'path': '/shares/{share_id}/metadata',
            }
        ]),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'delete_share_metadata',
        check_str=base.RULE_DEFAULT,
        description=("Delete share metadata."),
        operations=[
            {
                'method': 'DELETE',
                'path': '/shares/{share_id}/metadata/{key}',
            }
        ]),
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME % 'get_share_metadata',
        check_str=base.RULE_DEFAULT,
        description=("Get share metadata."),
        operations=[
            {
                'method': 'GET',
                'path': '/shares/{share_id}/metadata',
            }
        ]),
]


def list_rules():
    return shares_policies



# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from eventlet import greenthread
import mock
from oslo_concurrency import processutils
from six.moves.urllib import error as url_error
from six.moves.urllib import request as url_request

from manila import exception
from manila.share import configuration as conf
from manila.share.drivers.dell_emc.common.enas import connector
from manila import test
from manila.tests.share.drivers.dell_emc.common.enas import fakes
from manila.tests.share.drivers.dell_emc.common.enas import utils as enas_utils
from manila import utils


class XMLAPIConnectorTestData(object):
    FAKE_BODY = '<fakebody></fakebody>'
    FAKE_RESP = '<Response></Response>'
    FAKE_METHOD = 'fake_method'

    FAKE_KEY = 'key'
    FAKE_VALUE = 'value'

    @staticmethod
    def req_auth_url():
        return 'https://' + fakes.FakeData.emc_nas_server + '/Login'

    @staticmethod
    def req_credential():
        return (
            'user=' + fakes.FakeData.emc_nas_login
            + '&password=' + fakes.FakeData.emc_nas_password
            + '&Login=Login'
        ).encode()

    @staticmethod
    def req_url_encode():
        return {'Content-Type': 'application/x-www-form-urlencoded'}

    @staticmethod
    def req_url():
        return (
            'https://'
            + fakes.FakeData.emc_nas_server
            + '/servlets/CelerraManagementServices'
        )


XML_CONN_TD = XMLAPIConnectorTestData


class XMLAPIConnectorTest(test.TestCase):
    @mock.patch.object(url_request, 'Request', mock.Mock())
    def setUp(self):
        super(XMLAPIConnectorTest, self).setUp()

        emc_share_driver = fakes.FakeEMCShareDriver()

        self.configuration = emc_share_driver.configuration

        xml_socket = mock.Mock()
        xml_socket.read = mock.Mock(return_value=XML_CONN_TD.FAKE_RESP)
        opener = mock.Mock()
        opener.open = mock.Mock(return_value=xml_socket)

        with mock.patch.object(url_request, 'build_opener',
                               mock.Mock(return_value=opener)):
            self.XmlConnector = connector.XMLAPIConnector(
                configuration=self.configuration, debug=False)

            expected_calls = [
                mock.call(XML_CONN_TD.req_auth_url(),
                          XML_CONN_TD.req_credential(),
                          XML_CONN_TD.req_url_encode()),
            ]

            url_request.Request.assert_has_calls(expected_calls)

    def test_request_with_debug(self):
        self.XmlConnector.debug = True

        request = mock.Mock()
        request.headers = {XML_CONN_TD.FAKE_KEY: XML_CONN_TD.FAKE_VALUE}
        request.get_full_url = mock.Mock(
            return_value=XML_CONN_TD.FAKE_VALUE)

        with mock.patch.object(url_request, 'Request',
                               mock.Mock(return_value=request)):
            rsp = self.XmlConnector.request(XML_CONN_TD.FAKE_BODY,
                                            XML_CONN_TD.FAKE_METHOD)

            self.assertEqual(XML_CONN_TD.FAKE_RESP, rsp)

    def test_request_with_no_authorized_exception(self):
        xml_socket = mock.Mock()
        xml_socket.read = mock.Mock(return_value=XML_CONN_TD.FAKE_RESP)

        hook = enas_utils.RequestSideEffect()
        hook.append(ex=url_error.HTTPError(XML_CONN_TD.req_url(),
                                           '403', 'fake_message', None, None))
        hook.append(xml_socket)
        hook.append(xml_socket)

        self.XmlConnector.url_opener.open = mock.Mock(side_effect=hook)

        self.XmlConnector.request(XML_CONN_TD.FAKE_BODY)

    def test_request_with_general_exception(self):
        hook = enas_utils.RequestSideEffect()
        hook.append(ex=url_error.HTTPError(XML_CONN_TD.req_url(),
                                           'error_code', 'fake_message',
                                           None, None))
        self.XmlConnector.url_opener.open = mock.Mock(side_effect=hook)

        self.assertRaises(exception.ManilaException,
                          self.XmlConnector.request,
                          XML_CONN_TD.FAKE_BODY)


class MockSSH(object):
        def __enter__(self):
            return self

        def __exit__(self, type, value, traceback):
            pass


class MockSSHPool(object):
    def __init__(self):
        self.ssh = MockSSH()

    def item(self):
        try:
            return self.ssh
        finally:
            pass


class CmdConnectorTest(test.TestCase):
    def setUp(self):
        super(CmdConnectorTest, self).setUp()

        self.configuration = conf.Configuration(None)
        self.configuration.append_config_values = mock.Mock(return_value=0)
        self.configuration.emc_nas_login = fakes.FakeData.emc_nas_login
        self.configuration.emc_nas_password = fakes.FakeData.emc_nas_password
        self.configuration.emc_nas_server = fakes.FakeData.emc_nas_server
        self.configuration.emc_ssl_cert_verify = False
        self.configuration.emc_ssl_cert_path = None

        self.sshpool = MockSSHPool()
        with mock.patch.object(utils, "SSHPool",
                               mock.Mock(return_value=self.sshpool)):
            self.CmdHelper = connector.SSHConnector(
                configuration=self.configuration, debug=False)

            utils.SSHPool.assert_called_once_with(
                ip=fakes.FakeData.emc_nas_server,
                port=22,
                conn_timeout=None,
                login=fakes.FakeData.emc_nas_login,
                password=fakes.FakeData.emc_nas_password)

    def test_run_ssh(self):
        with mock.patch.object(processutils, "ssh_execute",
                               mock.Mock(return_value=('fake_output', ''))):
            cmd_list = ['fake', 'cmd']
            self.CmdHelper.run_ssh(cmd_list)

            processutils.ssh_execute.assert_called_once_with(
                self.sshpool.item(), 'fake cmd', check_exit_code=False)

    def test_run_ssh_with_debug(self):
        self.CmdHelper.debug = True

        with mock.patch.object(processutils, "ssh_execute",
                               mock.Mock(return_value=('fake_output', ''))):
            cmd_list = ['fake', 'cmd']
            self.CmdHelper.run_ssh(cmd_list)

            processutils.ssh_execute.assert_called_once_with(
                self.sshpool.item(), 'fake cmd', check_exit_code=False)

    @mock.patch.object(
        processutils, "ssh_execute",
        mock.Mock(side_effect=processutils.ProcessExecutionError))
    def test_run_ssh_exception(self):
        cmd_list = ['fake', 'cmd']

        self.mock_object(greenthread, 'sleep', mock.Mock())

        sshpool = MockSSHPool()

        with mock.patch.object(utils, "SSHPool",
                               mock.Mock(return_value=sshpool)):
            self.CmdHelper = connector.SSHConnector(self.configuration)

            self.assertRaises(processutils.ProcessExecutionError,
                              self.CmdHelper.run_ssh,
                              cmd_list,
                              True)

            utils.SSHPool.assert_called_once_with(
                ip=fakes.FakeData.emc_nas_server,
                port=22,
                conn_timeout=None,
                login=fakes.FakeData.emc_nas_login,
                password=fakes.FakeData.emc_nas_password)

            processutils.ssh_execute.assert_called_once_with(
                sshpool.item(), 'fake cmd', check_exit_code=True)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Add messages table

Revision ID: 238720805ce1
Revises: 31252d671ae5
Create Date: 2017-02-02 08:38:55.134095

"""

# revision identifiers, used by Alembic.
revision = '238720805ce1'
down_revision = '31252d671ae5'

from alembic import op
from oslo_log import log
from sqlalchemy import Column, DateTime
from sqlalchemy import MetaData, String, Table

LOG = log.getLogger(__name__)


def upgrade():
    meta = MetaData()
    meta.bind = op.get_bind()

    # New table
    messages = Table(
        'messages',
        meta,
        Column('id', String(36), primary_key=True, nullable=False),
        Column('project_id', String(255), nullable=False),
        Column('request_id', String(255), nullable=True),
        Column('resource_type', String(255)),
        Column('resource_id', String(36), nullable=True),
        Column('action_id', String(10), nullable=False),
        Column('detail_id', String(10), nullable=True),
        Column('message_level', String(255), nullable=False),
        Column('created_at', DateTime(timezone=False)),
        Column('updated_at', DateTime(timezone=False)),
        Column('deleted_at', DateTime(timezone=False)),
        Column('deleted', String(36)),
        Column('expires_at', DateTime(timezone=False)),
        mysql_engine='InnoDB',
        mysql_charset='utf8'
    )

    messages.create()


def downgrade():
    try:
        op.drop_table('messages')
    except Exception:
        LOG.error("messages table not dropped")
        raise



# Copyright 2013 OpenStack Foundation
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg
from oslo_utils import uuidutils

CONF = cfg.CONF


class FakeNetwork(object):
    def __init__(self, **kwargs):
        self.id = kwargs.pop('id', 'fake_net_id')
        self.name = kwargs.pop('name', 'net_name')
        self.subnets = kwargs.pop('subnets', [])
        for key, value in kwargs.items():
            setattr(self, key, value)

    def __getitem__(self, attr):
        return getattr(self, attr)


class FakeSubnet(object):
    def __init__(self, **kwargs):
        self.id = kwargs.pop('id', 'fake_subnet_id')
        self.network_id = kwargs.pop('network_id', 'fake_net_id')
        self.cidr = kwargs.pop('cidr', 'fake_cidr')
        for key, value in kwargs.items():
            setattr(self, key, value)

    def __getitem__(self, attr):
        return getattr(self, attr)


class FakePort(object):
    def __init__(self, **kwargs):
        self.id = kwargs.pop('id', 'fake_subnet_id')
        self.network_id = kwargs.pop('network_id', 'fake_net_id')
        self.fixed_ips = kwargs.pop('fixed_ips', [])
        for key, value in kwargs.items():
            setattr(self, key, value)

    def __getitem__(self, attr):
        return getattr(self, attr)


class FakeRouter(object):
    def __init__(self, **kwargs):
        self.id = kwargs.pop('id', 'fake_router_id')
        self.name = kwargs.pop('name', 'fake_router_name')
        for key, value in kwargs.items():
            setattr(self, key, value)

    def __getitem__(self, attr):
        return getattr(self, attr)

    def __setitem__(self, attr, value):
        setattr(self, attr, value)


class FakeDeviceAddr(object):
    def __init__(self, list_of_addresses=None):
        self.addresses = list_of_addresses or [
            dict(ip_version=4, cidr='1.0.0.0/27'),
            dict(ip_version=4, cidr='2.0.0.0/27'),
            dict(ip_version=6, cidr='3.0.0.0/27'),
        ]

    def list(self):
        return self.addresses


class FakeDevice(object):
    def __init__(self, name=None, list_of_addresses=None):
        self.addr = FakeDeviceAddr(list_of_addresses)
        self.name = name or 'fake_device_name'


class API(object):
    """Fake Network API."""
    admin_project_id = 'fake_admin_project_id'

    network = {
        "status": "ACTIVE",
        "subnets": ["fake_subnet_id"],
        "name": "fake_network",
        "tenant_id": "fake_tenant_id",
        "shared": False,
        "id": "fake_id",
        "router:external": False,
    }

    port = {
        "status": "ACTIVE",
        "allowed_address_pairs": [],
        "admin_state_up": True,
        "network_id": "fake_network_id",
        "tenant_id": "fake_tenant_id",
        "extra_dhcp_opts": [],
        "device_owner": "fake",
        "binding:capabilities": {"port_filter": True},
        "mac_address": "00:00:00:00:00:00",
        "fixed_ips": [
            {"subnet_id": "56537094-98d7-430a-b513-81c4dc6d9903",
             "ip_address": "10.12.12.10"}
        ],
        "id": "fake_port_id",
        "security_groups": ["fake_sec_group_id"],
        "device_id": "fake_device_id"
    }

    def get_all_admin_project_networks(self):
        net1 = self.network.copy()
        net1['tenant_id'] = self.admin_project_id
        net1['id'] = uuidutils.generate_uuid()

        net2 = self.network.copy()
        net2['tenant_id'] = self.admin_project_id
        net2['id'] = uuidutils.generate_uuid()
        return [net1, net2]

    def create_port(self, tenant_id, network_id, subnet_id=None,
                    fixed_ip=None, device_owner=None, device_id=None):
        port = self.port.copy()
        port['network_id'] = network_id
        port['admin_state_up'] = True
        port['tenant_id'] = tenant_id
        if fixed_ip:
            fixed_ip_dict = {'ip_address': fixed_ip}
            if subnet_id:
                fixed_ip_dict.update({'subnet_id': subnet_id})
            port['fixed_ips'] = [fixed_ip_dict]
        if device_owner:
            port['device_owner'] = device_owner
        if device_id:
            port['device_id'] = device_id
        return port

    def list_ports(self, **search_opts):
        """List ports for the client based on search options."""
        ports = []
        for i in range(2):
            ports.append(self.port.copy())
        for port in ports:
            port['id'] = uuidutils.generate_uuid()
            for key, val in search_opts.items():
                port[key] = val
            if 'id' in search_opts:
                return ports
        return ports

    def show_port(self, port_id):
        """Return the port for the client given the port id."""
        port = self.port.copy()
        port['id'] = port_id
        return port

    def delete_port(self, port_id):
        pass

    def get_subnet(self, subnet_id):
        pass

    def subnet_create(self, *args, **kwargs):
        pass

    def router_add_interface(self, *args, **kwargs):
        pass

    def show_router(self, *args, **kwargs):
        pass

    def update_port_fixed_ips(self, *args, **kwargs):
        pass

    def router_remove_interface(self, *args, **kwargs):
        pass

    def update_subnet(self, *args, **kwargs):
        pass

    def get_all_networks(self):
        """Get all networks for client."""
        net1 = self.network.copy()
        net2 = self.network.copy()
        net1['id'] = uuidutils.generate_uuid()
        net2['id'] = uuidutils.generate_uuid()
        return [net1, net2]

    def get_network(self, network_uuid):
        """Get specific network for client."""
        network = self.network.copy()
        network['id'] = network_uuid
        return network

    def network_create(self, tenant_id, name):
        network = self.network.copy()
        network['tenant_id'] = tenant_id
        network['name'] = name
        return network



# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ddt
from oslo_serialization import jsonutils as json
import requests
import requests_mock
import six

from manila import exception
from manila.share.drivers.dell_emc.plugins.isilon import isilon_api
from manila import test


@ddt.ddt
class IsilonApiTest(test.TestCase):

    def setUp(self):
        super(IsilonApiTest, self).setUp()

        self._mock_url = 'https://localhost:8080'
        _mock_auth = ('admin', 'admin')
        self.isilon_api = isilon_api.IsilonApi(
            self._mock_url, _mock_auth
        )

    @ddt.data(False, True)
    def test_create_directory(self, is_recursive):
        with requests_mock.Mocker() as m:
            path = '/ifs/test'
            self.assertEqual(0, len(m.request_history))
            self._add_create_directory_response(m, path, is_recursive)

            r = self.isilon_api.create_directory(path,
                                                 recursive=is_recursive)

            self.assertTrue(r)
            self.assertEqual(1, len(m.request_history))
            request = m.request_history[0]
            self._verify_dir_creation_request(request, path, is_recursive)

    @requests_mock.mock()
    def test_clone_snapshot(self, m):
        snapshot_name = 'snapshot01'
        fq_target_dir = '/ifs/admin/target'

        self.assertEqual(0, len(m.request_history))
        self._add_create_directory_response(m, fq_target_dir, False)
        snapshots_json = (
            '{"snapshots": '
            '[{"name": "snapshot01", "path": "/ifs/admin/source"}]'
            '}'
        )
        self._add_get_snapshot_response(m, snapshot_name, snapshots_json)

        # In order to test cloning a snapshot, we build out a mock
        # source directory tree. After the method under test is called we
        # will verify the necessary calls are made to clone a snapshot.
        source_dir_listing_json = (
            '{"children": ['
            '{"name": "dir1", "type": "container"},'
            '{"name": "dir2", "type": "container"},'
            '{"name": "file1", "type": "object"},'
            '{"name": "file2", "type": "object"}'
            ']}'
        )
        self._add_get_directory_listing_response(
            m, '/ifs/.snapshot/{0}/admin/source'.format(snapshot_name),
            source_dir_listing_json)

        # Add request responses for creating directories and cloning files
        # to the destination tree
        self._add_file_clone_response(m, '/ifs/admin/target/file1',
                                      snapshot_name)
        self._add_file_clone_response(m, '/ifs/admin/target/file2',
                                      snapshot_name)
        self._add_create_directory_response(m, fq_target_dir + '/dir1', False)
        self._add_get_directory_listing_response(
            m, '/ifs/.snapshot/{0}/admin/source/dir1'.format(snapshot_name),
            '{"children": ['
            '{"name": "file11", "type": "object"}, '
            '{"name": "file12", "type": "object"}'
            ']}')
        self._add_file_clone_response(m, '/ifs/admin/target/dir1/file11',
                                      snapshot_name)
        self._add_file_clone_response(m, '/ifs/admin/target/dir1/file12',
                                      snapshot_name)
        self._add_create_directory_response(m, fq_target_dir + '/dir2', False)
        self._add_get_directory_listing_response(
            m, '/ifs/.snapshot/{0}/admin/source/dir2'.format(snapshot_name),
            '{"children": ['
            '{"name": "file21", "type": "object"}, '
            '{"name": "file22", "type": "object"}'
            ']}')
        self._add_file_clone_response(m, '/ifs/admin/target/dir2/file21',
                                      snapshot_name)
        self._add_file_clone_response(m, '/ifs/admin/target/dir2/file22',
                                      snapshot_name)

        # Call method under test
        self.isilon_api.clone_snapshot(snapshot_name, fq_target_dir)

        # Verify calls needed to clone the source snapshot to the target dir
        expected_calls = []
        clone_path_list = [
            'file1', 'file2', 'dir1/file11', 'dir1/file12',
            'dir2/file21', 'dir2/file22']
        for path in clone_path_list:
            expected_call = IsilonApiTest.ExpectedCall(
                IsilonApiTest.ExpectedCall.FILE_CLONE,
                self._mock_url + '/namespace/ifs/admin/target/' + path,
                ['/ifs/admin/target/' + path, '/ifs/admin/source/' + path,
                 snapshot_name])
            expected_calls.append(expected_call)
        dir_path_list = [
            ('/dir1?recursive', '/dir1'),
            ('/dir2?recursive', '/dir2'),
            ('?recursive=', '')]
        for url, path in dir_path_list:
            expected_call = IsilonApiTest.ExpectedCall(
                IsilonApiTest.ExpectedCall.DIR_CREATION,
                self._mock_url + '/namespace/ifs/admin/target' + url,
                ['/ifs/admin/target' + path, False])
            expected_calls.append(expected_call)

        self._verify_clone_snapshot_calls(expected_calls, m.request_history)

    class ExpectedCall(object):
        DIR_CREATION = 'dir_creation'
        FILE_CLONE = 'file_clone'

        def __init__(self, request_type, match_url, verify_args):
            self.request_type = request_type
            self.match_url = match_url
            self.verify_args = verify_args

    def _verify_clone_snapshot_calls(self, expected_calls, response_calls):
        actual_calls = []
        for call in response_calls:
            actual_calls.append(call)
        for expected_call in expected_calls:
            # Match the expected call to the actual call, then verify
            match_found = False
            for call in actual_calls:
                if call.url.startswith(expected_call.match_url):
                    match_found = True
                    if expected_call.request_type is 'dir_creation':
                        self._verify_dir_creation_request(
                            call, *expected_call.verify_args)
                    elif expected_call.request_type is 'file_clone':
                        pass
                    else:
                        self.fail('Invalid request type')
                    actual_calls.remove(call)
            self.assertTrue(match_found)

    @requests_mock.mock()
    def test_get_directory_listing(self, m):
        self.assertEqual(0, len(m.request_history))
        fq_dir_path = 'ifs/admin/test'
        json_str = '{"my_json": "test123"}'
        self._add_get_directory_listing_response(m, fq_dir_path, json_str)

        actual_json = self.isilon_api.get_directory_listing(fq_dir_path)

        self.assertEqual(1, len(m.request_history))
        self.assertEqual(json.loads(json_str), actual_json)

    @ddt.data((200, True), (404, False))
    def test_is_path_existent(self, data):
        status_code, expected_return_value = data
        with requests_mock.mock() as m:
            self.assertEqual(0, len(m.request_history))
            path = '/ifs/home/admin'
            m.head('{0}/namespace{1}'.format(self._mock_url, path),
                   status_code=status_code)

            r = self.isilon_api.is_path_existent(path)

            self.assertEqual(expected_return_value, r)
            self.assertEqual(1, len(m.request_history))

    @requests_mock.mock()
    def test_is_path_existent_unexpected_error(self, m):
        path = '/ifs/home/admin'
        m.head('{0}/namespace{1}'.format(self._mock_url, path),
               status_code=400)

        self.assertRaises(
            requests.exceptions.HTTPError, self.isilon_api.is_path_existent,
            '/ifs/home/admin')

    @ddt.data(
        (200, '{"snapshots": [{"path": "/ifs/home/test"}]}',
         {'path': '/ifs/home/test'}),
        (404, '{"errors": []}', None)
    )
    def test_get_snapshot(self, data):
        status_code, json_body, expected_return_value = data
        with requests_mock.mock() as m:
            self.assertEqual(0, len(m.request_history))
            snapshot_name = 'foo1'
            self._add_get_snapshot_response(m, snapshot_name, json_body,
                                            status=status_code)

            r = self.isilon_api.get_snapshot(snapshot_name)

            self.assertEqual(1, len(m.request_history))
            self.assertEqual(expected_return_value, r)

    @requests_mock.mock()
    def test_get_snapshot_unexpected_error(self, m):
        snapshot_name = 'foo1'
        json_body = '{"snapshots": [{"path": "/ifs/home/test"}]}'
        self._add_get_snapshot_response(
            m, snapshot_name, json_body, status=400)

        self.assertRaises(
            requests.exceptions.HTTPError, self.isilon_api.get_snapshot,
            snapshot_name)

    @requests_mock.mock()
    def test_get_snapshots(self, m):
        self.assertEqual(0, len(m.request_history))
        snapshot_json = '{"snapshots": [{"path": "/ifs/home/test"}]}'
        m.get('{0}/platform/1/snapshot/snapshots'.format(self._mock_url),
              status_code=200, json=json.loads(snapshot_json))

        r = self.isilon_api.get_snapshots()

        self.assertEqual(1, len(m.request_history))
        self.assertEqual(json.loads(snapshot_json), r)

    @requests_mock.mock()
    def test_get_snapshots_error_occurred(self, m):
        self.assertEqual(0, len(m.request_history))
        m.get('{0}/platform/1/snapshot/snapshots'.format(self._mock_url),
              status_code=404)

        self.assertRaises(requests.exceptions.HTTPError,
                          self.isilon_api.get_snapshots)

        self.assertEqual(1, len(m.request_history))

    @ddt.data(
        ('/ifs/home/admin',
         '{"exports": [{"id": 42, "paths": ["/ifs/home/admin"]}]}', 42),
        ('/ifs/home/test',
         '{"exports": [{"id": 42, "paths": ["/ifs/home/admin"]}]}', None)
    )
    def test_lookup_nfs_export(self, data):
        share_path, response_json, expected_return = data
        with requests_mock.mock() as m:
            self.assertEqual(0, len(m.request_history))
            m.get('{0}/platform/1/protocols/nfs/exports'
                  .format(self._mock_url), json=json.loads(response_json))

            r = self.isilon_api.lookup_nfs_export(share_path)

            self.assertEqual(1, len(m.request_history))
            self.assertEqual(expected_return, r)

    @requests_mock.mock()
    def test_get_nfs_export(self, m):
        self.assertEqual(0, len(m.request_history))
        export_id = 42
        response_json = '{"exports": [{"id": 1}]}'
        status_code = 200
        m.get('{0}/platform/1/protocols/nfs/exports/{1}'
              .format(self._mock_url, export_id),
              json=json.loads(response_json), status_code=status_code)

        r = self.isilon_api.get_nfs_export(export_id)

        self.assertEqual(1, len(m.request_history))
        self.assertEqual(json.loads('{"id": 1}'), r)

    @requests_mock.mock()
    def test_get_nfs_export_error(self, m):
        self.assertEqual(0, len(m.request_history))
        export_id = 3
        response_json = '{}'
        status_code = 404
        m.get('{0}/platform/1/protocols/nfs/exports/{1}'
              .format(self._mock_url, export_id),
              json=json.loads(response_json), status_code=status_code)

        r = self.isilon_api.get_nfs_export(export_id)

        self.assertEqual(1, len(m.request_history))
        self.assertIsNone(r)

    @requests_mock.mock()
    def test_lookup_smb_share(self, m):
        self.assertEqual(0, len(m.request_history))
        share_name = 'my_smb_share'
        share_json = '{"id": "my_smb_share"}'
        response_json = '{{"shares": [{0}]}}'.format(share_json)
        m.get('{0}/platform/1/protocols/smb/shares/{1}'
              .format(self._mock_url, share_name), status_code=200,
              json=json.loads(response_json))

        r = self.isilon_api.lookup_smb_share(share_name)

        self.assertEqual(1, len(m.request_history))
        self.assertEqual(json.loads(share_json), r)

    @requests_mock.mock()
    def test_lookup_smb_share_error(self, m):
        self.assertEqual(0, len(m.request_history))
        share_name = 'my_smb_share'
        m.get('{0}/platform/1/protocols/smb/shares/{1}'.format(
            self._mock_url, share_name), status_code=404)

        r = self.isilon_api.lookup_smb_share(share_name)

        self.assertEqual(1, len(m.request_history))
        self.assertIsNone(r)

    @ddt.data((201, True), (404, False))
    def test_create_nfs_export(self, data):
        status_code, expected_return_value = data
        with requests_mock.mock() as m:
            self.assertEqual(0, len(m.request_history))
            export_path = '/ifs/home/test'
            m.post(self._mock_url + '/platform/1/protocols/nfs/exports',
                   status_code=status_code)

            r = self.isilon_api.create_nfs_export(export_path)

            self.assertEqual(1, len(m.request_history))
            call = m.request_history[0]
            expected_request_body = '{"paths": ["/ifs/home/test"]}'
            self.assertEqual(json.loads(expected_request_body),
                             json.loads(call.body))
            self.assertEqual(expected_return_value, r)

    @ddt.data((201, True), (404, False))
    def test_create_smb_share(self, data):
        status_code, expected_return_value = data
        with requests_mock.mock() as m:
            self.assertEqual(0, len(m.request_history))
            share_name = 'my_smb_share'
            share_path = '/ifs/home/admin/smb_share'
            m.post(self._mock_url + '/platform/1/protocols/smb/shares',
                   status_code=status_code)

            r = self.isilon_api.create_smb_share(share_name, share_path)

            self.assertEqual(expected_return_value, r)
            self.assertEqual(1, len(m.request_history))
            expected_request_data = {
                'name': share_name,
                'path': share_path,
                'permissions': []
            }
            self.assertEqual(expected_request_data,
                             json.loads(m.request_history[0].body))

    @requests_mock.mock()
    def test_create_snapshot(self, m):
        self.assertEqual(0, len(m.request_history))
        snapshot_name = 'my_snapshot_01'
        snapshot_path = '/ifs/home/admin'
        m.post(self._mock_url + '/platform/1/snapshot/snapshots',
               status_code=201)

        r = self.isilon_api.create_snapshot(snapshot_name, snapshot_path)

        self.assertEqual(1, len(m.request_history))
        self.assertTrue(r)
        expected_request_body = json.loads(
            '{{"name": "{0}", "path": "{1}"}}'
            .format(snapshot_name, snapshot_path)
        )
        self.assertEqual(expected_request_body,
                         json.loads(m.request_history[0].body))

    @requests_mock.mock()
    def test_create_snapshot_error_case(self, m):
        self.assertEqual(0, len(m.request_history))
        snapshot_name = 'my_snapshot_01'
        snapshot_path = '/ifs/home/admin'
        m.post(self._mock_url + '/platform/1/snapshot/snapshots',
               status_code=404)

        self.assertRaises(requests.exceptions.HTTPError,
                          self.isilon_api.create_snapshot,
                          snapshot_name, snapshot_path)
        self.assertEqual(1, len(m.request_history))

    @ddt.data(True, False)
    def test_delete(self, is_recursive_delete):
        with requests_mock.mock() as m:
            self.assertEqual(0, len(m.request_history))
            fq_path = '/ifs/home/admin/test'
            m.delete(self._mock_url + '/namespace' + fq_path + '?recursive='
                     + six.text_type(is_recursive_delete), status_code=204)

            self.isilon_api.delete(fq_path, recursive=is_recursive_delete)

            self.assertEqual(1, len(m.request_history))

    @requests_mock.mock()
    def test_delete_error_case(self, m):
        fq_path = '/ifs/home/admin/test'
        m.delete(self._mock_url + '/namespace' + fq_path + '?recursive=False',
                 status_code=403)

        self.assertRaises(requests.exceptions.HTTPError,
                          self.isilon_api.delete, fq_path, recursive=False)

    @ddt.data((204, True), (404, False))
    def test_delete_nfs_share(self, data):
        status_code, expected_return_value = data
        with requests_mock.mock() as m:
            self.assertEqual(0, len(m.request_history))
            share_number = 42
            m.delete('{0}/platform/1/protocols/nfs/exports/{1}'
                     .format(self._mock_url, share_number),
                     status_code=status_code)

            r = self.isilon_api.delete_nfs_share(share_number)

            self.assertEqual(1, len(m.request_history))
            self.assertEqual(expected_return_value, r)

    @ddt.data((204, True), (404, False))
    def test_delete_smb_shares(self, data):
        status_code, expected_return_value = data
        with requests_mock.mock() as m:
            self.assertEqual(0, len(m.request_history))

            share_name = 'smb_share_42'
            m.delete('{0}/platform/1/protocols/smb/shares/{1}'
                     .format(self._mock_url, share_name),
                     status_code=status_code)

            r = self.isilon_api.delete_smb_share(share_name)

            self.assertEqual(1, len(m.request_history))
            self.assertEqual(expected_return_value, r)

    @requests_mock.mock()
    def test_delete_snapshot(self, m):
        self.assertEqual(0, len(m.request_history))
        m.delete(self._mock_url + '/platform/1/snapshot/snapshots/my_snapshot',
                 status_code=204)

        self.isilon_api.delete_snapshot("my_snapshot")

        self.assertEqual(1, len(m.request_history))

    @requests_mock.mock()
    def test_delete_snapshot_error_case(self, m):
        m.delete(self._mock_url + '/platform/1/snapshot/snapshots/my_snapshot',
                 status_code=403)

        self.assertRaises(requests.exceptions.HTTPError,
                          self.isilon_api.delete_snapshot, "my_snapshot")

    @requests_mock.mock()
    def test_quota_create(self, m):
        quota_path = '/ifs/manila/test'
        quota_size = 256
        self.assertEqual(0, len(m.request_history))
        m.post(self._mock_url + '/platform/1/quota/quotas', status_code=201)

        self.isilon_api.quota_create(quota_path, 'directory', quota_size)

        self.assertEqual(1, len(m.request_history))
        expected_request_json = {
            'path': quota_path,
            'type': 'directory',
            'include_snapshots': False,
            'thresholds_include_overhead': False,
            'enforced': True,
            'thresholds': {'hard': quota_size},
        }
        call_body = m.request_history[0].body
        self.assertEqual(expected_request_json, json.loads(call_body))

    @requests_mock.mock()
    def test_quota_create__path_does_not_exist(self, m):
        quota_path = '/ifs/test2'
        self.assertEqual(0, len(m.request_history))
        m.post(self._mock_url + '/platform/1/quota/quotas', status_code=400)

        self.assertRaises(
            requests.exceptions.HTTPError,
            self.isilon_api.quota_create,
            quota_path, 'directory', 2
        )

    @requests_mock.mock()
    def test_quota_get(self, m):
        self.assertEqual(0, len(m.request_history))
        response_json = {'quotas': [{}]}
        m.get(self._mock_url + '/platform/1/quota/quotas', json=response_json,
              status_code=200)
        quota_path = "/ifs/manila/test"
        quota_type = "directory"

        self.isilon_api.quota_get(quota_path, quota_type)

        self.assertEqual(1, len(m.request_history))
        request_query_string = m.request_history[0].qs
        expected_query_string = {'path': [quota_path]}
        self.assertEqual(expected_query_string, request_query_string)

    @requests_mock.mock()
    def test_quota_get__path_does_not_exist(self, m):
        self.assertEqual(0, len(m.request_history))
        m.get(self._mock_url + '/platform/1/quota/quotas', status_code=404)

        response = self.isilon_api.quota_get(
            '/ifs/does_not_exist', 'directory')

        self.assertIsNone(response)

    @requests_mock.mock()
    def test_quota_modify(self, m):
        self.assertEqual(0, len(m.request_history))
        quota_id = "ADEF1G"
        new_size = 1024
        m.put('{0}/platform/1/quota/quotas/{1}'.format(
            self._mock_url, quota_id), status_code=204)

        self.isilon_api.quota_modify_size(quota_id, new_size)

        self.assertEqual(1, len(m.request_history))
        expected_request_body = {'thresholds': {'hard': new_size}}
        request_body = m.request_history[0].body
        self.assertEqual(expected_request_body, json.loads(request_body))

    @requests_mock.mock()
    def test_quota_modify__given_id_does_not_exist(self, m):
        quota_id = 'ADE2F'
        m.put('{0}/platform/1/quota/quotas/{1}'.format(
            self._mock_url, quota_id), status_code=404)

        self.assertRaises(
            requests.exceptions.HTTPError,
            self.isilon_api.quota_modify_size,
            quota_id, 1024
        )

    @requests_mock.mock()
    def test_quota_set__quota_already_exists(self, m):
        self.assertEqual(0, len(m.request_history))
        quota_path = '/ifs/manila/test'
        quota_type = 'directory'
        quota_size = 256
        quota_id = 'AFE2C'
        m.get('{0}/platform/1/quota/quotas'.format(
            self._mock_url), json={'quotas': [{'id': quota_id}]},
            status_code=200)
        m.put(
            '{0}/platform/1/quota/quotas/{1}'.format(self._mock_url, quota_id),
            status_code=204
        )

        self.isilon_api.quota_set(quota_path, quota_type, quota_size)

        expected_quota_modify_json = {'thresholds': {'hard': quota_size}}
        quota_put_json = json.loads(m.request_history[1].body)
        self.assertEqual(expected_quota_modify_json, quota_put_json)

    @requests_mock.mock()
    def test_quota_set__quota_does_not_already_exist(self, m):
        self.assertEqual(0, len(m.request_history))
        m.get('{0}/platform/1/quota/quotas'.format(
            self._mock_url), status_code=404)
        m.post('{0}/platform/1/quota/quotas'.format(self._mock_url),
               status_code=201)
        quota_path = '/ifs/manila/test'
        quota_type = 'directory'
        quota_size = 256

        self.isilon_api.quota_set(quota_path, quota_type, quota_size)

        # verify a call is made to create a quota
        expected_create_json = {
            six.text_type('path'): quota_path,
            six.text_type('type'): 'directory',
            six.text_type('include_snapshots'): False,
            six.text_type('thresholds_include_overhead'): False,
            six.text_type('enforced'): True,
            six.text_type('thresholds'): {six.text_type('hard'): quota_size},
        }
        create_request_json = json.loads(m.request_history[1].body)
        self.assertEqual(expected_create_json, create_request_json)

    @requests_mock.mock()
    def test_quota_set__path_does_not_already_exist(self, m):
        m.get(self._mock_url + '/platform/1/quota/quotas', status_code=400)

        e = self.assertRaises(
            requests.exceptions.HTTPError,
            self.isilon_api.quota_set,
            '/ifs/does_not_exist', 'directory', 2048
        )
        self.assertEqual(400, e.response.status_code)

    @ddt.data(
        ('foouser', isilon_api.SmbPermission.rw),
        ('testuser', isilon_api.SmbPermission.ro),
    )
    def test_smb_permission_add(self, data):
        user, smb_permission = data
        share_name = 'testshare'

        with requests_mock.mock() as m:
            papi_share_url = '{0}/platform/1/protocols/smb/shares/{1}'.format(
                self._mock_url, share_name)
            share_data = {
                'shares': [
                    {'permissions': []}
                ]
            }
            m.get(papi_share_url, status_code=200, json=share_data)

            auth_url = ('{0}/platform/1/auth/mapping/users/lookup?user={1}'
                        ''.format(self._mock_url, user))
            example_sid = 'SID:S-1-5-21'
            sid_json = {
                'id': example_sid,
                'name': user,
                'type': 'user'
            }
            auth_json = {'mapping': [
                {'user': {'sid': sid_json}}
            ]}
            m.get(auth_url, status_code=200, json=auth_json)
            m.put(papi_share_url)

            self.isilon_api.smb_permissions_add(share_name, user,
                                                smb_permission)

            perms_put_request = m.request_history[2]
            expected_perm_request_json = {
                'permissions': [
                    {'permission': smb_permission.value,
                     'permission_type': 'allow',
                     'trustee': sid_json
                     }
                ]
            }
            self.assertEqual(expected_perm_request_json,
                             json.loads(perms_put_request.body))

    @requests_mock.mock()
    def test_smb_permission_add_with_multiple_users_found(self, m):
        user = 'foouser'
        smb_permission = isilon_api.SmbPermission.rw
        share_name = 'testshare'
        papi_share_url = '{0}/platform/1/protocols/smb/shares/{1}'.format(
            self._mock_url, share_name)
        share_data = {
            'shares': [
                {'permissions': []}
            ]
        }
        m.get(papi_share_url, status_code=200, json=share_data)

        auth_url = ('{0}/platform/1/auth/mapping/users/lookup?user={1}'
                    ''.format(self._mock_url, user))
        example_sid = 'SID:S-1-5-21'
        sid_json = {
            'id': example_sid,
            'name': user,
            'type': 'user'
        }
        auth_json = {'mapping': [
            {'user': {'sid': sid_json}},
            {'user': {'sid': sid_json}},
        ]}
        m.get(auth_url, status_code=200, json=auth_json)
        m.put(papi_share_url)

        self.assertRaises(exception.ShareBackendException,
                          self.isilon_api.smb_permissions_add,
                          share_name, user, smb_permission)

    @requests_mock.mock()
    def test_smb_permission_remove(self, m):

        share_name = 'testshare'
        user = 'testuser'

        share_data = {
            'permissions': [{
                'permission': 'change',
                'permission_type': 'allow',
                'trustee': {
                    'id': 'SID:S-1-5-21',
                    'name': user,
                    'type': 'user',
                }
            }]
        }
        papi_share_url = '{0}/platform/1/protocols/smb/shares/{1}'.format(
            self._mock_url, share_name)
        m.get(papi_share_url, status_code=200, json={'shares': [share_data]})
        num_existing_perms = len(self.isilon_api.lookup_smb_share(share_name))
        self.assertEqual(1, num_existing_perms)

        m.put(papi_share_url)
        self.isilon_api.smb_permissions_remove(share_name, user)

        smb_put_request = m.request_history[2]
        expected_body = {'permissions': []}
        expected_body = json.dumps(expected_body)
        self.assertEqual(expected_body, smb_put_request.body)

    @requests_mock.mock()
    def test_smb_permission_remove_with_multiple_existing_perms(self, m):

        share_name = 'testshare'
        user = 'testuser'

        foouser_perms = {
            'permission': 'change',
            'permission_type': 'allow',
            'trustee': {
                'id': 'SID:S-1-5-21',
                'name': 'foouser',
                'type': 'user',
            }
        }
        user_perms = {
            'permission': 'change',
            'permission_type': 'allow',
            'trustee': {
                'id': 'SID:S-1-5-22',
                'name': user,
                'type': 'user',
            }
        }
        share_data = {
            'permissions': [
                foouser_perms,
                user_perms,
            ]
        }
        papi_share_url = '{0}/platform/1/protocols/smb/shares/{1}'.format(
            self._mock_url, share_name)
        m.get(papi_share_url, status_code=200, json={'shares': [share_data]})
        num_existing_perms = len(self.isilon_api.lookup_smb_share(
            share_name)['permissions'])
        self.assertEqual(2, num_existing_perms)
        m.put(papi_share_url)

        self.isilon_api.smb_permissions_remove(share_name, user)

        smb_put_request = m.request_history[2]
        expected_body = {'permissions': [foouser_perms]}
        expected_body = json.dumps(expected_body)
        self.assertEqual(json.loads(expected_body),
                         json.loads(smb_put_request.body))

    @requests_mock.mock()
    def test_smb_permission_remove_with_empty_perms_list(self, m):
        share_name = 'testshare'
        user = 'testuser'

        share_data = {'permissions': []}
        papi_share_url = '{0}/platform/1/protocols/smb/shares/{1}'.format(
            self._mock_url, share_name)
        m.get(papi_share_url, status_code=200, json={'shares': [share_data]})
        m.put(papi_share_url)

        self.assertRaises(exception.ShareBackendException,
                          self.isilon_api.smb_permissions_remove,
                          share_name, user)

    @requests_mock.mock()
    def test_auth_lookup_user(self, m):
        user = 'foo'
        auth_url = '{0}/platform/1/auth/mapping/users/lookup?user={1}'.format(
            self._mock_url, user)
        example_sid = 'SID:S-1-5-21'
        sid_json = {
            'id': example_sid,
            'name': user,
            'type': 'user'
        }
        auth_json = {
            'mapping': [
                {'user': {'sid': sid_json}}
            ]
        }
        m.get(auth_url, status_code=200, json=auth_json)

        returned_auth_json = self.isilon_api.auth_lookup_user(user)
        self.assertEqual(auth_json, returned_auth_json)

    @requests_mock.mock()
    def test_auth_lookup_user_with_nonexistent_user(self, m):
        user = 'nonexistent'
        auth_url = '{0}/platform/1/auth/mapping/users/lookup?user={1}'.format(
            self._mock_url, user)
        m.get(auth_url, status_code=404)
        self.assertRaises(exception.ShareBackendException,
                          self.isilon_api.auth_lookup_user, user)

    @requests_mock.mock()
    def test_auth_lookup_user_with_backend_error(self, m):
        user = 'foo'
        auth_url = '{0}/platform/1/auth/mapping/users/lookup?user={1}'.format(
            self._mock_url, user)
        m.get(auth_url, status_code=400)
        self.assertRaises(requests.exceptions.HTTPError,
                          self.isilon_api.auth_lookup_user, user)

    def _add_create_directory_response(self, m, path, is_recursive):
        url = '{0}/namespace{1}?recursive={2}'.format(
            self._mock_url, path, six.text_type(is_recursive))
        m.put(url, status_code=200)

    def _add_file_clone_response(self, m, fq_dest_path, snapshot_name):
        url = '{0}/namespace{1}?clone=true&snapshot={2}'.format(
            self._mock_url, fq_dest_path, snapshot_name)
        m.put(url)

    def _add_get_directory_listing_response(self, m, fq_dir_path, json_str):
        url = '{0}/namespace{1}?detail=default'.format(
            self._mock_url, fq_dir_path)
        m.get(url, json=json.loads(json_str), status_code=200)

    def _add_get_snapshot_response(
            self, m, snapshot_name, json_str, status=200):
        url = '{0}/platform/1/snapshot/snapshots/{1}'.format(
            self._mock_url, snapshot_name
        )
        m.get(url, status_code=status, json=json.loads(json_str))

    def _verify_dir_creation_request(self, request, path, is_recursive):
        self.assertEqual('PUT', request.method)
        expected_url = '{0}/namespace{1}?recursive={2}'.format(
            self._mock_url, path, six.text_type(is_recursive))
        self.assertEqual(expected_url, request.url)
        self.assertIn("x-isi-ifs-target-type", request.headers)
        self.assertEqual("container",
                         request.headers['x-isi-ifs-target-type'])

    def _verify_clone_file_from_snapshot(
            self, request, fq_file_path, fq_dest_path, snapshot_name):
        self.assertEqual('PUT', request.method)
        expected_url = '{0}/namespace{1}?clone=true&snapshot={2}'.format(
            self._mock_url, fq_dest_path, snapshot_name
        )
        self.assertEqual(expected_url, request.request.url)
        self.assertIn("x-isi-ifs-copy-source", request.headers)
        self.assertEqual('/namespace' + fq_file_path,
                         request.headers['x-isi-ifs-copy-source'])



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ddt
import mock
from webob import exc

from manila.api.v2 import share_replica_export_locations as export_locations
from manila.common import constants
from manila import context
from manila import db
from manila import exception
from manila import policy
from manila import test
from manila.tests.api import fakes
from manila.tests import db_utils


@ddt.ddt
class ShareReplicaExportLocationsAPITest(test.TestCase):

    def _get_request(self, version="2.47", use_admin_context=False):
        req = fakes.HTTPRequest.blank(
            '/v2/share-replicas/%s/export-locations' % self.active_replica_id,
            version=version, use_admin_context=use_admin_context,
            experimental=True)
        return req

    def setUp(self):
        super(ShareReplicaExportLocationsAPITest, self).setUp()
        self.controller = (
            export_locations.ShareReplicaExportLocationController())
        self.resource_name = 'share_replica_export_location'
        self.ctxt = context.RequestContext('fake', 'fake')
        self.mock_policy_check = self.mock_object(
            policy, 'check_policy', mock.Mock(return_value=True))
        self.share = db_utils.create_share(
            replication_type=constants.REPLICATION_TYPE_READABLE,
            replica_state=constants.REPLICA_STATE_ACTIVE)
        self.active_replica_id = self.share.instance.id
        self.req = self._get_request()
        exports = [
            {'path': 'myshare.mydomain/active-replica-exp1',
             'is_admin_only': False},
            {'path': 'myshare.mydomain/active-replica-exp2',
             'is_admin_only': False},
        ]
        db.share_export_locations_update(
            self.ctxt, self.active_replica_id, exports)

        # Replicas
        self.share_replica2 = db_utils.create_share_replica(
            share_id=self.share.id,
            replica_state=constants.REPLICA_STATE_IN_SYNC)
        self.share_replica3 = db_utils.create_share_replica(
            share_id=self.share.id,
            replica_state=constants.REPLICA_STATE_OUT_OF_SYNC)
        replica2_exports = [
            {'path': 'myshare.mydomain/insync-replica-exp',
             'is_admin_only': False},
            {'path': 'myshare.mydomain/insync-replica-exp2',
             'is_admin_only': False}
        ]
        replica3_exports = [
            {'path': 'myshare.mydomain/outofsync-replica-exp',
             'is_admin_only': False},
            {'path': 'myshare.mydomain/outofsync-replica-exp2',
             'is_admin_only': False}
        ]
        db.share_export_locations_update(
            self.ctxt, self.share_replica2.id, replica2_exports)
        db.share_export_locations_update(
            self.ctxt, self.share_replica3.id, replica3_exports)

    @ddt.data('user', 'admin')
    def test_list_and_show(self, role):
        summary_keys = [
            'id', 'path', 'replica_state', 'availability_zone', 'preferred'
        ]
        admin_summary_keys = summary_keys + [
            'share_instance_id', 'is_admin_only'
        ]
        detail_keys = summary_keys + ['created_at', 'updated_at']
        admin_detail_keys = admin_summary_keys + ['created_at', 'updated_at']

        self._test_list_and_show(role, summary_keys, detail_keys,
                                 admin_summary_keys, admin_detail_keys)

    def _test_list_and_show(self, role, summary_keys, detail_keys,
                            admin_summary_keys, admin_detail_keys):

        req = self._get_request(use_admin_context=(role == 'admin'))
        for replica_id in (self.active_replica_id, self.share_replica2.id,
                           self.share_replica3.id):
            index_result = self.controller.index(req, replica_id)

            self.assertIn('export_locations', index_result)
            self.assertEqual(1, len(index_result))
            self.assertEqual(2, len(index_result['export_locations']))

            for index_el in index_result['export_locations']:
                self.assertIn('id', index_el)
                show_result = self.controller.show(
                    req, replica_id, index_el['id'])
                self.assertIn('export_location', show_result)
                self.assertEqual(1, len(show_result))

                show_el = show_result['export_location']

                # Check summary keys in index result & detail keys in show
                if role == 'admin':
                    self.assertEqual(len(admin_summary_keys), len(index_el))
                    for key in admin_summary_keys:
                        self.assertIn(key, index_el)
                    self.assertEqual(len(admin_detail_keys), len(show_el))
                    for key in admin_detail_keys:
                        self.assertIn(key, show_el)
                else:
                    self.assertEqual(len(summary_keys), len(index_el))
                    for key in summary_keys:
                        self.assertIn(key, index_el)
                    self.assertEqual(len(detail_keys), len(show_el))
                    for key in detail_keys:
                        self.assertIn(key, show_el)

                # Ensure keys common to index & show have matching values
                for key in summary_keys:
                    self.assertEqual(index_el[key], show_el[key])

    def test_list_and_show_with_non_replicas(self):
        non_replicated_share = db_utils.create_share()
        instance_id = non_replicated_share.instance.id
        exports = [
            {'path': 'myshare.mydomain/non-replicated-share',
             'is_admin_only': False},
            {'path': 'myshare.mydomain/non-replicated-share-2',
             'is_admin_only': False},
        ]
        db.share_export_locations_update(self.ctxt, instance_id, exports)
        updated_exports = db.share_export_locations_get_by_share_id(
            self.ctxt, non_replicated_share.id)

        self.assertRaises(exc.HTTPNotFound, self.controller.index, self.req,
                          instance_id)

        for export in updated_exports:
            self.assertRaises(exc.HTTPNotFound, self.controller.show, self.req,
                              instance_id, export['id'])

    def test_list_export_locations_share_replica_not_found(self):
        self.assertRaises(
            exc.HTTPNotFound,
            self.controller.index,
            self.req, 'non-existent-share-replica-id')

    def test_show_export_location_share_replica_not_found(self):
        index_result = self.controller.index(self.req, self.active_replica_id)
        el_id = index_result['export_locations'][0]['id']

        self.assertRaises(
            exc.HTTPNotFound,
            self.controller.show,
            self.req, 'non-existent-share-replica-id', el_id)

        self.assertRaises(
            exc.HTTPNotFound,
            self.controller.show,
            self.req, self.active_replica_id,
            'non-existent-export-location-id')

    @ddt.data('1.0', '2.0', '2.46')
    def test_list_with_unsupported_version(self, version):
        self.assertRaises(
            exception.VersionNotFoundForAPIMethod,
            self.controller.index,
            self._get_request(version),
            self.active_replica_id)

    @ddt.data('1.0', '2.0', '2.46')
    def test_show_with_unsupported_version(self, version):
        index_result = self.controller.index(self.req, self.active_replica_id)

        self.assertRaises(
            exception.VersionNotFoundForAPIMethod,
            self.controller.show,
            self._get_request(version),
            self.active_replica_id,
            index_result['export_locations'][0]['id'])
